/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SWO_DEBUG	FALSE
#define FE_SWO_VERBOSE	FALSE

#include "operator/operator.pmh"

using namespace fe;

void SurfaceWrapOp::initialize(void)
{
	//* NOTE inherits attributes from SurfaceBindOp

	catalogRemove("InputGroup");
	catalogRemove("FragmentAttr");
	catalogRemove("DriverGroup");
	catalogRemove("Bindings");
	catalogRemove("Bind Primitives");
	catalogRemove("Bind Fragments");
	catalogRemove("Covert");
	catalogRemove("Conform");
	catalogRemove("LimitDistance");
	catalogRemove("MaxDistance");
	catalogRemove("PartitionDriver");
	catalogRemove("InputPartitionAttr");
	catalogRemove("DriverPartitionAttr");
	catalogRemove("PivotPoint");
	catalogRemove("Scalable");

	catalog<String>("BindPrefix","hint")=
			"Prefix of attribute names on input surface describing"
			" where each point on the input surface is attached to"
			" the associated driver.";

	catalog<bool>("DiscardBindings")=true;
	catalog<String>("DiscardBindings","label")="Discard Bindings";
	catalog<String>("DiscardBindings","page")="Config";
	catalog<String>("DiscardBindings","hint")=
			"Remove used binding attributes from the output.";

	catalog<bool>("checkDeformed")=false;
	catalog<String>("checkDeformed","label")="Check Deformed";
	catalog<String>("checkDeformed","page")="Config";
	catalog<bool>("checkDeformed","joined")=true;
	catalog<String>("checkDeformed","hint")=
			"Verify that the deformed driver surface is like"
			" the reference driver surface."
			"  These checks take time, so this option should probably"
			" only be used temporarily.";

	catalog<I32>("checkLimit")=1;
	catalog<I32>("checkLimit","min")=0;
	catalog<I32>("checkLimit","high")=10;
	catalog<I32>("checkLimit","max")=100;
	catalog<String>("checkLimit","enabler")="checkDeformed";
	catalog<String>("checkLimit","label")="Message Limit";
	catalog<String>("checkLimit","page")="Config";
	catalog<bool>("checkLimit","joined")=true;
	catalog<String>("checkLimit","hint")=
			"Stop reporting mismatches after the given number of errors.";

	catalog<Real>("scaleLimit")=2;
	catalog<Real>("scaleLimit","min")=1;
	catalog<Real>("scaleLimit","high")=10;
	catalog<String>("scaleLimit","enabler")="checkDeformed";
	catalog<String>("scaleLimit","label")="Scale Limit";
	catalog<String>("scaleLimit","page")="Config";
	catalog<String>("scaleLimit","hint")=
			"Threshold for rescale warnings on deformed driver.";

	catalog<Real>("Distance Bias")=1.0;
	catalog<Real>("Distance Bias","min")=0.0;
	catalog<Real>("Distance Bias","high")=10.0;
	catalog<Real>("Distance Bias","max")=1e3;
	catalog<String>("Distance Bias","page")="Bind";
	catalog<bool>("Distance Bias","joined")=true;
	catalog<String>("Distance Bias","hint")=
			"Dominance of closer faces towards weighting";

	catalog<Real>("Distance Scale")=1.0;
	catalog<Real>("Distance Scale","min")=0.0;
	catalog<Real>("Distance Scale","high")=2.0;
	catalog<Real>("Distance Scale","max")=100.0;
	catalog<String>("Distance Scale","page")="Bind";
	catalog<String>("Distance Scale","hint")=
			"Distance gain for weighting";

	catalog<String>("SampleMethod")="linear";
	catalog<String>("SampleMethod","label")="Sampling";
	catalog<String>("SampleMethod","choice:0")="linear";
	catalog<String>("SampleMethod","label:0")="Linear";
	catalog<String>("SampleMethod","choice:1")="pointNormal";
	catalog<String>("SampleMethod","label:1")="Point Normal";
	catalog<String>("SampleMethod","choice:2")="spline";
	catalog<String>("SampleMethod","label:2")="Spline";
	catalog<String>("SampleMethod","page")="Bind";
	catalog<bool>("SampleMethod","joined")=true;
	catalog<String>("SampleMethod","hint")=
			"Type of interpolation between driver vertices."
			"  Point Normal takes about four times longer than Linear."
			"  Spline currently only works on curves.";

	catalog<bool>("rotateNormals")=false;
	catalog<String>("rotateNormals","label")="Rotate Normals";
	catalog<String>("rotateNormals","page")="Bind";
	catalog<String>("rotateNormals","hint")=
			"Apply transforms to normals in addition to positions.";

	catalog<bool>("spread")=false;
	catalog<String>("spread","label")="Spread";
	catalog<String>("spread","page")="Bind";
	catalog<bool>("spread","joined")=true;
	catalog<String>("spread","hint")=
			"Apportion weights onto neighboring triangles"
					" using barycentric coordinates.";

	catalog<I32>("spreadDepth")=1;
	catalog<String>("spreadDepth","label")="Spread Depth";
	catalog<I32>("spreadDepth","min")=1;
	catalog<I32>("spreadDepth","high")=10;
	catalog<I32>("spreadDepth","max")=100;
	catalog<String>("spreadDepth","page")="Bind";
	catalog<String>("spreadDepth","enabler")="spread";
	catalog<String>("spreadDepth","hint")=
			"Number of iterations to propagate weights.";

	catalog<bool>("singleTransform")=false;
	catalog<String>("singleTransform","label")="Single Transform";
	catalog<String>("singleTransform","page")="Bind";
	catalog<bool>("singleTransform","joined")=true;
	catalog<String>("singleTransform","hint")=
			"Use a single weighted blend of all binding transforms."
			"  Each binding simply translates in that space.";

	catalog<Real>("singleBias")=10.0;
	catalog<String>("singleBias","label")="Singularity";
	catalog<String>("singleBias","page")="Bind";
	catalog<Real>("singleBias","min")=0.0;
	catalog<Real>("singleBias","high")=20.0;
	catalog<Real>("singleBias","max")=100.0;
	catalog<String>("singleBias","enabler")="singleTransform";
	catalog<String>("singleBias","hint")=
			"Dominance of closer faces towards weighting";

	catalog<bool>("Debug Bind")=false;
	catalog<bool>("Debug Bind","joined")=true;
	catalog<String>("Debug Bind","page")="Bind";
	catalog<String>("Debug Bind","hint")=
			"Draws lines over the output showing the attachments."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	catalog<I32>("Debug Index")=-1;
	catalog<I32>("Debug Index","min")=-1;
	catalog<I32>("Debug Index","high")=1000;
	catalog<I32>("Debug Index","max")=1000000;
	catalog<String>("Debug Index","page")="Bind";
	catalog<String>("Debug Index","enabler")="Debug Bind";
	catalog<String>("Debug Index","hint")=
			"Index of element to examine (-1 to see them all).";

	catalog<bool>("test0")=false;
	catalog<String>("test0","label")="test0: no single transform influences";
	catalog<String>("test0","page")="Bind";

	catalog<bool>("test1")=false;
	catalog<String>("test1","label")="test1: no single transform multi-weight";
	catalog<String>("test1","page")="Bind";

	catalog<bool>("test2")=false;
	catalog<String>("test2","label")="test2: use distance to triangle center";
	catalog<String>("test2","page")="Bind";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")=
			"Click an output point to see its bindings.";

	catalog<String>("Input Surface","label")="Bound Input Surface";

	catalog< sp<Component> >("Deformed Surface");
	catalog<String>("Deformed Surface","label")="Deformed Driver Surface";

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";

	//* check problems
	catalog<bool>("Output Surface","recycle")=true;

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");

	m_spBrush=new DrawMode();
	m_spBrush->setDrawStyle(DrawMode::e_foreshadow);
	m_spBrush->setLineWidth(2.0);
	m_spBrush->setAntialias(TRUE);
	m_spBrush->setLit(FALSE);

	m_spOverlay=new DrawMode();
	m_spOverlay->setLineWidth(3.0);
	m_spOverlay->setLit(FALSE);
}

void SurfaceWrapOp::handle(Record& a_rSignal)
{
	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle brush %d brushed %d recook %d\n",
			spDrawBrush.isValid(),m_brushed,m_forceRecook);
#endif

	if(spDrawBrush.isValid())
	{
		m_brushed=TRUE;

		const Color transparent(0.0,0.0,0.0,0.0);
		const Color black(0.0,0.0,0.0);
		const Color white(1.0,1.0,0.8);
		const Color grey(0.8,0.8,1.0);
		const Color red(0.8,0.0,0.0);
		const Color green(0.2,1.0,0.2);
		const Color blue(0.0,0.0,1.0);
		const Color cyan(0.0,0.8,1.0);
		const Color magenta(0.8,0.0,0.8);
		const Color yellow(0.8,0.8,0.0);

		const Real dotRadius=6.0;
		const Real rimSize=6.0;
		const Real maxAlpha=0.6;
		const Real fadeAlpha=0.5;

		if(m_spOutput.isValid())
		{
			sp<SurfaceAccessorI> spOutputVertices;
			sp<SurfaceAccessorI> spOutputPoints;
			if(m_spOutputAccessible.isValid())
			{
				if(!access(spOutputVertices,m_spOutputAccessible,
						e_primitive,e_vertices)) return;
				if(!access(spOutputPoints,m_spOutputAccessible,
						e_point,e_position)) return;
			}

			sp<DrawI> spDrawOverlay;
			accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

			spDrawBrush->pushDrawMode(m_spBrush);
			spDrawOverlay->pushDrawMode(m_spOverlay);

			sp<ViewI> spView=spDrawOverlay->view();

			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			m_event.bind(windowEvent(a_rSignal));

			if(!m_event.isExpose())
			{
				const BWORD leftPress=
						m_event.isMousePress(WindowEvent::e_itemLeft);
				if(leftPress)
				{
					m_forceRecook=TRUE;
				}

				I32 pickIndex= -1;

				const Real maxDistance=100.0;
				sp<SurfaceSearchable::Impact> spImpact=m_spOutput->rayImpact(
						rRayOrigin,rRayDirection,maxDistance);
				if(spImpact.isValid())
				{
					const I32 primitiveIndex=spImpact->primitiveIndex();
					I32 pointIndex= -1;

					sp<SurfaceTriangles::Impact> spTriImpact=spImpact;
					if(spTriImpact.isValid())
					{
						pointIndex=nearestPointIndex(spTriImpact);
					}

					sp<SurfaceCurves::Impact> spCurveImpact=spImpact;
					if(spCurveImpact.isValid())
					{
						pointIndex=nearestPointIndex(
								spOutputVertices,spCurveImpact);
					}

					if(pointIndex>=0)
					{
						pickIndex=pointIndex;
					}

//					feLog("prim %d point %d\n",primitiveIndex,pointIndex);
				}

				if(pickIndex>=0)
				{
					if(leftPress)
					{
						m_selectIndex=pickIndex;
					}

					if(pickIndex!=m_selectIndex)
					{
						const SpatialVector point=
								spOutputPoints->spatialVector(pickIndex);

						drawDot(spView,spDrawOverlay,point,dotRadius,blue);
					}
				}
				else if(leftPress)
				{
					m_selectIndex=-1;
				}
			}

			if(m_selectIndex>=0)
			{
				const SpatialVector point=spOutputPoints->spatialVector(
						m_selectIndex);

				drawSquare(spDrawOverlay,point,rimSize,
						black,cyan,NULL,NULL);

				Real maxWeight=1e-9;
				Real minDist=0.0;
				Real maxDist=0.0;

				const U32 feedbackCount=m_feedback.size();
				Real distanceArray[feedbackCount];

				for(U32 feedbackIndex=0;feedbackIndex<feedbackCount;
						feedbackIndex++)
				{
					const Feedback& rFeedback=m_feedback[feedbackIndex];
					const Real weight=rFeedback.weight();

					if(maxWeight<weight)
					{
						maxWeight=weight;
					}

					sp<SurfaceI::ImpactI> spImpact=rFeedback.impact();
					const SpatialVector point=spImpact->intersection();
					const Real viewDist=magnitude(point-rRayOrigin);

					if(!feedbackIndex || minDist>viewDist)
					{
						minDist=viewDist;
					}
					if(maxDist<viewDist)
					{
						maxDist=viewDist;
					}

					distanceArray[feedbackIndex]=viewDist;
				}

				const Real deltaDistance=
						(maxDist>minDist)? maxDist-minDist: 1.0;

				for(U32 feedbackIndex=0;feedbackIndex<feedbackCount;
						feedbackIndex++)
				{
					const Feedback& rFeedback=m_feedback[feedbackIndex];
					const Real weight=rFeedback.weight();
					sp<SurfaceI::ImpactI> spImpact=rFeedback.impact();

					const Real viewDist=distanceArray[feedbackIndex];
					const Real alpha=maxAlpha-
							fadeAlpha*(viewDist-minDist)/deltaDistance;

					const Color weighted(1.0,weight/maxWeight,0.0,alpha);
					spDrawBrush->draw(spImpact,&weighted);

#if FALSE
					const U32 margin=1;
					Vector2i projected=spView->project(
							spImpact->intersection(),
							ViewMode::e_perspective);
					String text;
					text.sPrintf("%.3f",weight);
					drawLabel(spDrawOverlay,projected,text,
							FALSE,margin,cyan,NULL,&black);
#endif
				}
			}

			spDrawOverlay->popDrawMode();
			spDrawBrush->popDrawMode();
		}

		return;
	}

	const Real frame=currentFrame(a_rSignal);
	const BWORD frameChanged=(frame!= m_lastFrame);
	const BWORD replaced=(catalog<bool>("Output Surface","replaced"));
	const BWORD paramChanged=(!m_brushed && !replaced && !frameChanged);

	const BWORD rebuild=
			(m_forceRecook || paramChanged || frameChanged || replaced);

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle frame %.6G->%.6G paramChanged %d"
			" frameChanged %d replaced %d forceRecook %d -> rebuild %d\n",
			m_lastFrame,frame,paramChanged,frameChanged,replaced,
			m_forceRecook,rebuild);
#endif

	m_lastFrame=frame;
	m_forceRecook=FALSE;
	m_brushed=FALSE;

	if(!rebuild)
	{
		//* brush doesn't change output

		return;
	}

	//* output curves is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	if(catalog<BWORD>("Debug Bind"))
	{
		if(!accessGuide(m_spDrawDebug,a_rSignal)) return;
	}
	else
	{
		m_spDrawDebug=NULL;
	}

	if(!access(m_spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessibleI> spDriverAccessible;
	if(!access(spDriverAccessible,"Driver Surface")) return;

	sp<SurfaceAccessorI> spDriverVertices;
	if(!access(spDriverVertices,spDriverAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spDeformedVertices;
	if(!access(spDeformedVertices,"Deformed Surface",
			e_primitive,e_vertices)) return;

	const String driverSignature=generateSignature(spDriverAccessible);
	const String inputSignature=retrieveSignature(m_spInputAccessible);
	if(driverSignature!=inputSignature)
	{
		String message;
		message.sPrintf("Reference driver surface has a different signature"
				" than when bound (\"%s\" vs \"%s\");",
				driverSignature.c_str(),inputSignature.c_str());
		catalog<String>("warning")+=message;
	}

	String driverGroup;
	String fragmentAttribute;
	String inputGroup;
	String covertKey;
	retrieveParameters(m_spInputAccessible,driverGroup,fragmentAttribute,
			inputGroup,m_bindCount,m_bindPrimitives,m_fragmented,covertKey);

	m_spBindingAccessible=m_spInputAccessible;
	if(!covertKey.empty())
	{
		sp<SurfaceAccessibleCatalog> spSurfaceAccessibleCatalog=
				registry()->create("*.SurfaceAccessibleCatalog");
		if(spSurfaceAccessibleCatalog.isValid())
		{
			spSurfaceAccessibleCatalog->setCatalog(
					registry()->master()->catalog());
			spSurfaceAccessibleCatalog->setKey(covertKey);
			m_spBindingAccessible=spSurfaceAccessibleCatalog;
		}
	}

	const BWORD checkDeformed=catalog<bool>("checkDeformed");
	const U32 driverPrimitiveCount=spDriverVertices->count();
	const U32 deformedPrimitiveCount=spDeformedVertices->count();
	if(deformedPrimitiveCount!=driverPrimitiveCount)
	{
		String message;
		message.sPrintf("Deformed driver surface has a different number"
				" of primitives than reference (%d vs %d);",
				deformedPrimitiveCount,driverPrimitiveCount);
		catalog<String>(checkDeformed? "error": "warning")+=message;
		if(checkDeformed)
		{
			return;
		}
	}
	else if(checkDeformed)
	{
		sp<SurfaceAccessorI> spDriverPoints;
		if(!access(spDriverPoints,spDriverAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spDeformedPoints;
		if(!access(spDeformedPoints,"Deformed Surface",
				e_point,e_position)) return;

		const I32 checkLimit=catalog<I32>("checkLimit");
		const Real scaleLimit=catalog<Real>("scaleLimit");
		const Real invScaleLimit=scaleLimit>0.0? 1.0/scaleLimit: 0.0;
		U32 warnings=0;
		U32 mismatches=0;
		for(U32 primitiveIndex=0;
				primitiveIndex<driverPrimitiveCount && mismatches<checkLimit;
				primitiveIndex++)
		{
			const U32 driverSubCount=
					spDriverVertices->subCount(primitiveIndex);
			const U32 deformedSubCount=
					spDeformedVertices->subCount(primitiveIndex);
			if(driverSubCount!=deformedSubCount)
			{
				String message;
				message.sPrintf("  primitive %d has different"
						" number of vertices (%d vs %d);\n",primitiveIndex,
						deformedPrimitiveCount,driverPrimitiveCount);
				catalog<String>("error")+=message;
				mismatches++;
			}
			else
			{
				BWORD broken=FALSE;
				SpatialVector lastDriverPoint;
				SpatialVector lastDeformedPoint;
				for(U32 subIndex=0;
						subIndex<driverSubCount &&
						warnings+mismatches<checkLimit;
						subIndex++)
				{
					const U32 driverPointIndex=spDriverVertices->integer(
							primitiveIndex,subIndex);
					const U32 deformedPointIndex=spDeformedVertices->integer(
							primitiveIndex,subIndex);
					const SpatialVector driverPoint=
							spDriverPoints->spatialVector(driverPointIndex);
					const SpatialVector deformedPoint=
							spDeformedPoints->spatialVector(deformedPointIndex);

					if(FE_INVALID_SCALAR(driverPoint[0]) ||
							FE_INVALID_SCALAR(driverPoint[1]) ||
							FE_INVALID_SCALAR(driverPoint[2]))
					{
						String message;
						message.sPrintf(
								"  driver primitive %d vertex %d"
								" has corrupt data;\n",
								primitiveIndex,subIndex);
						catalog<String>("error")+=message;
						mismatches++;
					}
					else if(FE_INVALID_SCALAR(deformedPoint[0]) ||
							FE_INVALID_SCALAR(deformedPoint[1]) ||
							FE_INVALID_SCALAR(deformedPoint[2]))
					{
						String message;
						message.sPrintf(
								"  deformed primitive %d vertex %d"
								" has corrupt data;\n",
								primitiveIndex,subIndex);
						catalog<String>("error")+=message;
						mismatches++;
					}
					else if(driverPointIndex!=deformedPointIndex)
					{
						String message;
						message.sPrintf("  primitive %d has mismatched"
								" vertex %d (%d vs %d);\n",
								primitiveIndex,subIndex,
								deformedPointIndex,driverPointIndex);
						catalog<String>("error")+=message;
						mismatches++;
					}
					else if(subIndex && !broken)
					{
						const Real driverLength=
								magnitude(driverPoint-lastDriverPoint);
						const Real deformedLength=
								magnitude(deformedPoint-lastDeformedPoint);

						if(driverLength<1e-6)
						{
							String message;
							message.sPrintf(
									"  driver primitive %d segment %d"
									" has near zero length;\n",
									primitiveIndex,subIndex-1);
							catalog<String>("error")+=message;
							mismatches++;
						}
						else if(deformedLength<1e-6)
						{
							String message;
							message.sPrintf(
									"  deformed primitive %d segment %d"
									" has near zero length;\n",
									primitiveIndex,subIndex-1);
							catalog<String>("error")+=message;
							mismatches++;
						}
						else
						{
							const Real segmentScale=deformedLength/driverLength;
							if(segmentScale<invScaleLimit ||
									segmentScale>scaleLimit)
							{
								String message;
								message.sPrintf(
										"  deformed primitive %d segment %d"
										" has scaled by %.6G\n",
										primitiveIndex,subIndex-1,
										deformedLength/driverLength);
								catalog<String>("warning")+=message;
								warnings++;
							}
						}
					}
					lastDriverPoint=driverPoint;
					lastDeformedPoint=deformedPoint;
				}
			}
		}
		if(mismatches)
		{
			String message;
			message.sPrintf("Deformed driver has %d%s mismatches:\n",
					mismatches,mismatches>=checkLimit? "+": "");
			catalog<String>("error")=message+catalog<String>("error");
			return;
		}
	}

	const SurfaceI::SampleMethod sampleMethod=
			(catalog<String>("SampleMethod")=="spline")?
			SurfaceI::e_spline:
			((catalog<String>("SampleMethod")=="pointNormal")?
			SurfaceI::e_pointNormal: SurfaceI::e_linear);

	//* original driver
	if(!access(m_spDriver,"Driver Surface",driverGroup)) return;

	m_spDriver->setRefinement(0);
	m_spDriver->setSampleMethod(sampleMethod);
	m_spDriver->prepareForSample();
	if(m_spDriver->radius()<1e-6)
	{
		catalog<String>("error")="Reference Driver Surface appears empty"
				" (zero radius bounding sphere);";
		return;
	}

	//* deformed driver
	if(!access(m_spDeformed,"Deformed Surface",driverGroup)) return;

	m_spDeformed->setRefinement(0);
	m_spDeformed->setSampleMethod(sampleMethod);
	m_spDeformed->prepareForSample();
	if(m_spDeformed->radius()<1e-6)
	{
		catalog<String>("error")="Deformed Driver Surface appears empty"
				" (zero radius bounding sphere);";
		return;
	}

	m_element=m_bindPrimitives? e_primitive: e_point;
	m_attribute=m_bindPrimitives? e_vertices: e_position;

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	if(catalog<bool>("rotateNormals"))
	{
		sp<SurfaceAccessorI> spInputNormals;
		sp<SurfaceAccessorI> spOutputNormals;

		if(!access(spInputNormals,m_spInputAccessible,
				e_point,e_normal,e_quiet) ||
				!access(spOutputNormals,m_spOutputAccessible,
				e_point,e_normal,e_quiet))
		{
			catalog<String>("warning")+="can not rotate missing normals;";
		}
		else
		{
			//* flush normals
			const U32 normalCount=spInputNormals->count();
			for(U32 normalIndex=0;normalIndex<normalCount;normalIndex++)
			{
				spOutputNormals->set(normalIndex,
						spInputNormals->spatialVector(normalIndex));
			}
		}
	}

	const U32 elementCount=spOutputElements->count();

	m_spread=catalog<bool>("spread");
	if(m_spread)
	{
		//* TEMP only allow driver triangles for spread

		sp<SurfaceAccessorI> spDriverVertices;
		if(access(spDriverVertices,"Driver Surface",e_primitive,e_vertices))
		{
			const U32 driverPrimCount=spDriverVertices->count();
			for(U32 driverPrimIndex=0;driverPrimIndex<driverPrimCount;
					driverPrimIndex++)
			{
				if(spDriverVertices->subCount(driverPrimIndex)!=3)
				{
					catalog<String>("warning")+="currently, spread is"
							" only allowed on triangulated drivers;";
					m_spread=FALSE;
					continue;
				}
			}
		}

		if(m_bindCount>1)
		{
			catalog<String>("warning")+=
					"currently, spread only uses single bindings;";
		}
	}

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->pushDrawMode(m_spDebugGroup);
	}

	U32 runCount=elementCount;
	m_spFragmentAccessor=NULL;
	if(m_fragmented)
	{
		if(access(m_spFragmentAccessor,m_spInputAccessible,
				m_element,m_attribute))
		{
			m_spFragmentAccessor->fragmentWith(
					SurfaceAccessibleI::Element(m_element),
					fragmentAttribute,inputGroup);
			runCount=m_spFragmentAccessor->fragmentCount();
			if(runCount<1)	//* was (runCount<2), but there can be one fragment
			{
				catalog<String>("warning")+=
						"fragmentation failed to divide surface -> reverting";
				m_fragmented=FALSE;
				runCount=elementCount;
			}
		}
		else
		{
			catalog<String>("warning")+=
					"fragmentation access failed -> ignoring";
			m_fragmented=FALSE;
			runCount=elementCount;
		}
	}

	m_feedback.clear();

	//* TODO fragment-aware ranges
	if(m_fragmented)
	{
		//* serial
		sp<SurfaceRange> spFullRange=sp<SurfaceRange>(new SurfaceRange());
		spFullRange->nonAtomic().add(0,runCount-1);
		run(-1,spFullRange);
	}
	else
	{
		setAtomicRange(m_spOutputAccessible,
				m_bindPrimitives? e_pointsOfPrimitives: e_pointsOnly);

		OperatorThreaded::handle(a_rSignal);
	}

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->popDrawMode();
	}

	if(catalog<bool>("DiscardBindings"))
	{
		discardBindings(m_spOutputAccessible,m_bindCount,m_bindPrimitives);
		discardSignature(m_spOutputAccessible);
		discardParameters(m_spOutputAccessible);
	}

	//* output for brush probing
	if(!access(m_spOutput,"Output Surface")) return;
	m_spOutput->setRefinement(0);

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle done\n");
#endif
}

void SurfaceWrapOp::run(I32 a_id,sp<SurfaceRange> a_spRange)
{
	const BWORD singleTransform=catalog<bool>("singleTransform");
	const Real singleBias=catalog<Real>("singleBias");
	const I32 spreadDepth=catalog<I32>("spreadDepth");
	const BWORD rotateNormals=catalog<bool>("rotateNormals");
	const Real distanceBias=catalog<Real>("Distance Bias");
	const Real distanceScale=catalog<Real>("Distance Scale");
	const I32 debugIndex=catalog<I32>("Debug Index");
	const BWORD test0=catalog<bool>("test0");
	const BWORD test1=catalog<bool>("test1");
	const BWORD test2=catalog<bool>("test2");

	const BWORD cacheOutput=FALSE;

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run running thread %d %s\n",
			a_id,a_spRange->brief().c_str());
#endif

	sp<SurfaceAccessorI> spInputElements;
	if(!access(spInputElements,m_spInputAccessible,
			m_element,m_attribute)) return;

	sp<SurfaceAccessorI> spOutputElements;
	if(cacheOutput)
	{
		sp<SurfaceAccessorI> spOutputElementsDirect;
		if(!access(spOutputElementsDirect,m_spOutputAccessible,
				m_element,m_attribute)) return;

		sp<SurfaceAccessorCached> spAccessorCached=
				registry()->create("*.SurfaceAccessorCached");
		FEASSERT(spAccessorCached.isValid());
		spAccessorCached->setSurfaceAccessible(m_spOutputAccessible);
		spAccessorCached->setAccessorChain(spOutputElementsDirect);

		spOutputElements=spAccessorCached;
	}
	else
	{
		if(!access(spOutputElements,m_spOutputAccessible,
				m_element,m_attribute)) return;
	}

	sp<SurfaceAccessorI> spInputNormals;
	sp<SurfaceAccessorI> spOutputNormals;
	if(rotateNormals)
	{
		access(spInputNormals,m_spInputAccessible,
				e_point,e_normal,e_quiet);
		access(spOutputNormals,m_spOutputAccessible,
				e_point,e_normal,e_quiet);
	}

	std::vector< sp<SurfaceAccessorI> > inputFaces;
	std::vector< sp<SurfaceAccessorI> > inputBarys;
	const I32 bindingsAccessed=accessBindings(m_spBindingAccessible,
			inputFaces,inputBarys,m_bindCount,m_bindPrimitives,FALSE);

	if(bindingsAccessed<1)
	{
		return;
	}

	const SpatialBary baryThird(0.33,0.33);

	sp<SurfaceAccessorI> spDriverVertices;
	if(!access(spDriverVertices,"Driver Surface",
			e_primitive,e_vertices)) return;
	const U32 driverPrimCount=spDriverVertices->count();

	std::vector< std::set<U32> > pointPrimitives;
	if(m_spread)
	{
		sp<SurfaceAccessorI> spInputPoints;
		if(!access(spInputPoints,m_spInputAccessible,e_point,
				e_position,e_quiet)) return;

		const U32 pointCount=spInputPoints->count();
		pointPrimitives.resize(pointCount);

		for(U32 primitiveIndex=0;primitiveIndex<driverPrimCount;
				primitiveIndex++)
		{
			const U32 subCount=spDriverVertices->subCount(primitiveIndex);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spDriverVertices->integer(primitiveIndex,subIndex);
				pointPrimitives[pointIndex].insert(primitiveIndex);
			}
		}
	}

	const BWORD singular=
			(m_fragmented && m_element==SurfaceAccessibleI::e_point);

	sp<SurfaceAccessorI::FilterI> spFilter;

	for(SurfaceRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 index=it.value();

		I32 filterCount=1;
		U32 tableCount=0;
		if(m_fragmented)
		{
			const String fragment=m_spFragmentAccessor->fragment(index);
			if(!m_spFragmentAccessor->filterWith(fragment,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}
			filterCount=spFilter->filterCount();
			if(filterCount<1)
			{
				continue;
			}
			if(singular)
			{
				tableCount+=filterCount;
			}
			else
			{
				for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
				{
					const I32 filtered=spFilter->filter(filterIndex);
					FEASSERT(filtered>=0);
					tableCount+=spInputElements->subCount(filtered);
				}
			}
		}
		else
		{
			tableCount=spInputElements->subCount(index);
		}

		if(tableCount<1)
		{
			continue;
		}

		SpatialVector point[tableCount];
		SpatialVector normal[tableCount];
		SpatialVector resultPoint[tableCount];
		SpatialVector resultNormal[tableCount];
		Real totalWeight[tableCount];

		I32 feedbackIndex= -1;
		U32 tableIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const I32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			FEASSERT(filtered>=0);
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=m_bindPrimitives?
						spInputElements->integer(filtered,subIndex): filtered;
				if(pointIndex==m_selectIndex)
				{
					feedbackIndex=tableIndex;
				}

				point[tableIndex]=
						spInputElements->spatialVector(filtered,subIndex);
				if(spInputNormals.isValid())
				{
					normal[tableIndex]=
							spInputNormals->spatialVector(pointIndex);
				}
				set(resultPoint[tableIndex]);
				set(resultNormal[tableIndex]);
				totalWeight[tableIndex]=0.0;
				tableIndex++;
			}
		}

		const I32 keyIndex=m_fragmented? spFilter->filter(0): index;
		FEASSERT(keyIndex>=0);

		//* spread Vector2i: <face, mask>
		std::vector<Vector2i> spreadData;
		std::vector<SpatialVector> spreadWeight;
		if(m_spread)
		{
			//* HACK presuming triangles only
			//* TODO quads

			const U32 binding=0;	//* TODO start with multiple bindings

			const I32 face=inputFaces[binding]->integer(keyIndex);
			const SpatialBary barycenter=
					inputBarys[binding]->spatialVector(keyIndex);

			spreadData.push_back(Vector2i(face,7));
			spreadWeight.push_back(SpatialVector(barycenter[0],barycenter[1],
						1.0-barycenter[0]-barycenter[1]));

			for(U32 spreadIndex=0;spreadIndex<spreadDepth;spreadIndex++)
			{
				//* accumulate weight per vertex
				std::map<U32,Real> weightMap;
				const U32 faceCount=spreadData.size();
				for(U32 faceIndex=0;faceIndex<faceCount;faceIndex++)
				{
					const I32 spreadFace=spreadData[faceIndex][0];
					const SpatialVector weights=spreadWeight[faceIndex];

					for(U32 subIndex=0;subIndex<3;subIndex++)
					{
						const U32 pointIndex=
								spDriverVertices->integer(spreadFace,subIndex);
						weightMap[pointIndex]+=weights[subIndex];
					}
				}

				spreadData.clear();
				spreadWeight.clear();

				for(std::map<U32,Real>::iterator it=weightMap.begin();
						it!=weightMap.end();it++)
				{
					const U32 pointIndex=it->first;
					const Real totalWeight=it->second;

					std::set<U32>& rPrimitiveSet=pointPrimitives[pointIndex];
					const U32 shares=rPrimitiveSet.size();
					FEASSERT(shares);

					const Real shareWeight=totalWeight/Real(shares);

					for(std::set<U32>::iterator it2=rPrimitiveSet.begin();
							it2!=rPrimitiveSet.end();it2++)
					{
						const U32 primitiveIndex= *it2;

						//* HACK
						const I32 spreadFace=primitiveIndex;

						const U32 faceCount=spreadData.size();
						U32 faceIndex;
						for(faceIndex=0;faceIndex<faceCount;faceIndex++)
						{
							const I32 otherFace=spreadData[faceIndex][0];

							if(otherFace==spreadFace)
							{
								spreadWeight[faceIndex]+=shareWeight*
										SpatialVector(1.0/3.0,1.0/3.0,1.0/3.0);
								break;
							}
						}

						if(faceIndex==faceCount)
						{
							spreadData.push_back(Vector2i(spreadFace,7));
							spreadWeight.push_back(shareWeight*
									SpatialVector(1.0/3.0,1.0/3.0,1.0/3.0));
						}
					}
				}
			}

			//* put original binding first
			const U32 faceCount=spreadData.size();
			for(U32 faceIndex=0;faceIndex<faceCount;faceIndex++)
			{
				const I32 spreadFace=spreadData[faceIndex][0];
				if(spreadFace==face)
				{
					if(faceIndex)
					{
						const Vector2i data=spreadData[faceIndex];
						const SpatialVector weights=spreadWeight[faceIndex];

						spreadData[faceIndex]=spreadData[0];
						spreadWeight[faceIndex]=spreadWeight[0];

						spreadData[0]=data;
						spreadWeight[0]=weights;
					}
					break;
				}
			}
		}

		Real peakDistance=0.0;
		std::vector<Real> bindingDistances;
		for(U32 binding=0;binding<bindingsAccessed;binding++)
		{
			const I32 face=inputFaces[binding]->integer(keyIndex);
			if(face<0)
			{
				bindingDistances.push_back(-1.0);
				continue;
			}

			//* test2: distance to center or nearest point
			const SpatialBary barycenter=(test2 && binding)? baryThird:
					SpatialBary(inputBarys[binding]->spatialVector(keyIndex));

			const SpatialVector to=
					m_spDriver->sample(face,barycenter).translation();

			const Real distance=magnitude(to-point[0]);

			bindingDistances.push_back(distance);
			if(peakDistance<distance)
			{
				peakDistance=distance;
			}
		}

		SpatialTransform coreConversion;
		if(singleTransform)
		{
			//* TODO cleanup

			set(coreConversion.column(0));
			set(coreConversion.column(1));
			set(coreConversion.column(2));
			set(coreConversion.column(3));

			Real totalWeight=0.0;
			for(U32 binding=0;binding<bindingsAccessed;binding++)
			{
				if(test0 && binding)
				{
					continue;
				}

				const I32 face=inputFaces[binding]->integer(keyIndex);
				if(face<0)
				{
					continue;
				}
				SpatialBary barycenter=
						inputBarys[binding]->spatialVector(keyIndex);

				SpatialTransform xformDriver=
						m_spDriver->sample(face,barycenter);
				SpatialTransform xformDeformed=
						m_spDeformed->sample(face,barycenter);

				SpatialTransform invDriver;
				invert(invDriver,xformDriver);
				SpatialTransform conversion=invDriver*xformDeformed;

				const Real distance=
						magnitude(xformDriver.translation()-point[0]);

				//* in case all points are equally distant
				const Real minWeight=binding? 0.0: 1e-3;

				const Real weight=calculateWeight(distance,peakDistance,
						singleBias,1.0,minWeight);

				coreConversion=coreConversion+weight*conversion;

				if(!test1 || !binding)
				{
					totalWeight+=weight;
				}
			}
			SpatialVector coreX=coreConversion.column(0);
			SpatialVector coreY=unitSafe(coreConversion.column(1));
			SpatialVector coreT=coreConversion.translation()*
					(totalWeight>0.0? 1.0/totalWeight: 1.0);
			makeFrameNormalY(coreConversion,coreT,coreX,coreY);
		}

		SpatialVector baryWeights;

		const U32 bindCount=m_spread? spreadData.size(): bindingsAccessed;
		for(U32 binding=0;binding<bindCount;binding++)
		{
			if(test1 && binding)
			{
				continue;
			}

			const I32 face=m_spread? spreadData[binding][0]:
					inputFaces[binding]->integer(keyIndex);
			if(face<0)
			{
				continue;
			}

			SpatialTransform conversion;
			SpatialVector translateDriver;
			SpatialVector translateDeformed;

			SpatialBary barycenter=baryThird;
			if(!m_spread || !binding)
			{
				barycenter=inputBarys[binding]->spatialVector(keyIndex);
			}
			else
			{
				barycenter.solve(
						spDriverVertices->spatialVector(face,0),
						spDriverVertices->spatialVector(face,1),
						spDriverVertices->spatialVector(face,2),
						point[0]);

				//* clamp to triangle
				barycenter.clamp(
						spDriverVertices->spatialVector(face,0),
						spDriverVertices->spatialVector(face,1),
						spDriverVertices->spatialVector(face,2));
			}

			if(m_spread)
			{
				//* TODO proportional to area

				baryWeights=spreadWeight[binding];
			}

#if FE_SWO_VERBOSE
			feLog("SurfaceWrapOp::handle element %d bind %d/%d\n",
					index,binding,bindingsAccessed);
#endif

			if(!singleTransform)
			{
				const SpatialTransform xformDriver=
						m_spDriver->sample(face,barycenter);
				const SpatialTransform xformDeformed=
						m_spDeformed->sample(face,barycenter);

#if FE_SWO_VERBOSE
				feLog("xformDriver\n%s\n",c_print(xformDriver));
				feLog("xformDeformed\n%s\n",c_print(xformDeformed));
#endif

				SpatialTransform invDriver;
				invert(invDriver,xformDriver);
				conversion=invDriver*xformDeformed;

				translateDriver=xformDriver.translation();
				translateDeformed=xformDeformed.translation();

#if FALSE
				feLog("xformDriver\n%s\n",c_print(xformDriver));
				feLog("xformDeformed\n%s\n",c_print(xformDeformed));
#endif
			}
			else
			{
				//* TODO sample translation only
				translateDriver=
						m_spDriver->sample(face,barycenter).translation();
				translateDeformed=
						m_spDeformed->sample(face,barycenter).translation();

				SpatialVector refDeformed;
				transformVector(coreConversion,translateDriver,refDeformed);

				conversion=coreConversion;
				conversion.translation()+=translateDeformed-refDeformed;
			}

			Real weight=0.0;

			if(m_spread)
			{
				const I32 mask=spreadData[binding][1];
				if(mask&1)
				{
					weight+=baryWeights[0];
				}
				if(mask&2)
				{
					weight+=baryWeights[1];
				}
				if(mask&4)
				{
					weight+=baryWeights[2];
				}
			}
			else if(bindingsAccessed<2)
			{
				weight=1.0;
			}
			else
			{
				const Real distance=bindingDistances[binding];

				//* in case all points are equally distant
				const Real minWeight=binding? 0.0: 1e-3;

				weight=calculateWeight(distance,peakDistance,
						distanceBias,distanceScale,minWeight);


//				feLog("%d bind %d/%d face %d dist %.6G/%.6G weight %.6G\n",
//						index,binding,bindingsAccessed,face,
//						distance,peakDistance,weight);
			}

			for(tableIndex=0;tableIndex<tableCount;tableIndex++)
			{
				if(I32(tableIndex)==feedbackIndex)
				{
					sp<SurfaceI::ImpactI> spImpact=
							m_spDeformed->sampleImpact(face,barycenter);
					if(spImpact.isValid())
					{
						m_feedback.push_back(Feedback(weight,spImpact));
					}
				}

				totalWeight[tableIndex]+=weight;

				SpatialVector transformedPoint;
				transformVector(conversion,point[tableIndex],transformedPoint);
				resultPoint[tableIndex]+=weight*transformedPoint;

				if(rotateNormals)
				{
					SpatialVector transformedNormal;
					rotateVector(conversion,normal[tableIndex],
							transformedNormal);
					resultNormal[tableIndex]+=weight*transformedNormal;
				}

#if FE_SWO_VERBOSE
				feLog(" point %d/%d face %d bary %s weight %.6G"
						" from %s to %s\n",
						tableIndex,tableCount,
						face,c_print(barycenter),weight,
						c_print(point[tableIndex]),c_print(transformedPoint));
				feLog("conversion\n%s\n",c_print(conversion));
				feLog("translateDriver\n%s\n",c_print(translateDriver));
				feLog("translateDeformed\n%s\n",c_print(translateDeformed));
#endif
			}

			if(m_spDrawDebug.isValid() && (debugIndex<0 || debugIndex==index))
			{
				SpatialVector transformedPoint;
				transformVector(conversion,point[0],transformedPoint);

				const Color yellow(1.0,1.0,0.0,1.0);
				SpatialVector line[2];

				line[0]=transformedPoint;
				line[1]=translateDeformed;

				SAFEGUARD;
				m_spDrawDebug->drawLines(line,NULL,2,DrawI::e_strip,
						false,&yellow);
			}
		}

		tableIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const I32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			FEASSERT(filtered>=0);
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				if(totalWeight[tableIndex]>0.0)
				{
					spOutputElements->set(filtered,subIndex,
							resultPoint[tableIndex]/totalWeight[tableIndex]);
					if(rotateNormals && spOutputNormals.isValid())
					{
						const I32 pointIndex=m_bindPrimitives?
								spOutputElements->integer(filtered,subIndex):
								filtered;

						spOutputNormals->set(pointIndex,
								unitSafe(resultNormal[tableIndex]));
					}
				}

				tableIndex++;
			}
		}
	}

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run done running thread %d\n",a_id);
#endif
}

Real SurfaceWrapOp::calculateWeight(Real a_distance,Real a_peakDistance,
	Real a_distanceBias,Real a_distanceScale,Real a_minimum)
{
//	Real weight=1.0/(1.0+pow(a_distanceScale*a_distance,a_distanceBias));

	Real weight=pow(fmax(0.0,1.0-a_distanceScale*a_distance/a_peakDistance),
			a_distanceBias);

	if(weight<a_minimum)
	{
		weight=a_minimum;
	}

	return weight;
}
