/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SC_SAMPLE_DEBUG	FALSE

#include "surface.pmh"

namespace fe
{

SurfaceCurves::SurfaceCurves(void)
{
	m_pImpactPool=&m_curveImpactPool;
}

Protectable* SurfaceCurves::clone(Protectable* pInstance)
{
	feLog("SurfaceCurves::clone\n");

	SurfaceCurves* pSurfaceCurve= pInstance?
			dynamic_cast<SurfaceCurves*>(pInstance): new SurfaceCurves();

	checkCache();

	//* TODO

	return pSurfaceCurve;
}

void SurfaceCurves::cache(void)
{
	if(!m_record.isValid())
	{
		feLog("SurfaceCurves::cache record invalid\n");
		return;
	}

	SurfaceSphere::cache();

	//* TODO

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(m_record);

	sp<RecordGroup> spRG=surfaceModelRV.surfacePointRG();
	for(RecordGroup::iterator it=spRG->begin();it!=spRG->end();it++)
	{
		sp<RecordArray> spRA= *it;

		SurfacePoint m_surfacePointRV;
		m_surfacePointRV.bind(spRA);
		while(m_surfacePointRV.step())
		{
			SpatialVector& at=m_surfacePointRV.at();
			SpatialVector& up=m_surfacePointRV.up();

			m_pVertexArray=(SpatialVector*)reallocate(m_pVertexArray,
					(m_vertices+1)*sizeof(SpatialVector));

			SpatialVector& rVertex=m_pVertexArray[m_vertices];

			rVertex=at;

			printf("vertex %d  %s\n",m_vertices,c_print(rVertex));
			m_vertices++;
		}
	}
}

sp<SurfaceI::ImpactI> SurfaceCurves::sampleImpact(I32 a_triangleIndex,
	SpatialBary a_barycenter,SpatialVector a_tangent) const
{
	//* TODO clean up duplicate code with sample()
	const I32 startIndex=m_pElementArray[a_triangleIndex][0];
	const I32 subCount=m_pElementArray[a_triangleIndex][1];
	if(startIndex+subCount>m_vertices)
	{
		return sp<SurfaceI::ImpactI>(NULL);
	}

	const Real realIndex=a_barycenter[0]*(subCount-1);
	const I32 subIndex=I32(realIndex>0.0?
			(realIndex<(subCount-1.0)? realIndex: subCount-2.0): 0.0);
	const I32 arrayIndex=startIndex+subIndex;

	const SpatialTransform xform=
			sample(a_triangleIndex,a_barycenter,a_tangent);

	//* TODO refactor with nearly identical code in SurfaceTriangles.cc
	sp<Impact> spImpact=m_pImpactPool->get();

	spImpact->setSurface(dynamic_cast<const SurfaceCurves*>(this));

	spImpact->setTriangleIndex(a_triangleIndex);
	spImpact->setBarycenter(a_barycenter);

	spImpact->setPointIndex0(m_pPointIndexArray[arrayIndex]);
	spImpact->setPointIndex1(m_pPointIndexArray[arrayIndex+1]);

	spImpact->setVertex0(m_pVertexArray[arrayIndex]);
	spImpact->setVertex1(m_pVertexArray[arrayIndex+1]);

	spImpact->setNormal0(m_pNormalArray[arrayIndex]);
	spImpact->setNormal1(m_pNormalArray[arrayIndex+1]);

	const SpatialVector intersection=xform.translation();
	const Real distance=magnitude(intersection);
	spImpact->setDistance(distance);
	spImpact->setDirection(intersection*(distance>0.0? 1.0/distance: 1.0));

	return spImpact;
}

//* in this context, a_triangleIndex is really the curve primitive index
//* TODO convert triangle logic to segment logic
SpatialTransform SurfaceCurves::sample(I32 a_triangleIndex,
	SpatialBary a_barycenter,SpatialVector a_tangent) const
{
#if FE_SC_SAMPLE_DEBUG
	feLog("SurfaceCurves::sample triangleIndex %d bary %s tan %s\n",
			a_triangleIndex,c_print(a_barycenter),c_print(a_tangent));
#endif

	const I32 startIndex=m_pElementArray[a_triangleIndex][0];
	const I32 subCount=m_pElementArray[a_triangleIndex][1];
	if(startIndex+subCount>m_vertices)
	{
#if FE_SC_SAMPLE_DEBUG
		//* WARNING multi-threaded logging can segfault
		feLog("SurfaceCurves::sample"
				" excess triangleIndex %d (%d>%d)\n",
				a_triangleIndex,startIndex+subCount,m_vertices);
#endif
		return SpatialTransform::identity();
	}

	const Real realIndex=a_barycenter[0]*(subCount-1);
	const I32 subIndex=I32(realIndex>0.0?
			(realIndex<(subCount-1.0)? realIndex: subCount-2.0): 0.0);
	const Real weight1=realIndex-subIndex;
	const Real weight0=1.0-weight1;

	const I32 arrayIndex=startIndex+subIndex;
	const I32 arrayIndex1=arrayIndex+1;

	const SpatialVector& rVertex0=m_pVertexArray[arrayIndex];
	const SpatialVector& rVertex1=m_pVertexArray[arrayIndex1];

	const SpatialVector& rNormal0=m_pNormalArray[arrayIndex];
	const SpatialVector& rNormal1=m_pNormalArray[arrayIndex1];

	SpatialVector point;
	SpatialVector norm;
	if(m_sampleMethod==e_spline)
	{
		const U32 arrayIndexM=(subIndex>0)? arrayIndex-1: arrayIndex;
		const U32 arrayIndex2=(subIndex<subCount-2)? arrayIndex+2: arrayIndex1;

		const SpatialVector& rVertexM=m_pVertexArray[arrayIndexM];
		const SpatialVector& rVertex2=m_pVertexArray[arrayIndex2];

		const SpatialVector& rNormalM=m_pNormalArray[arrayIndexM];
		const SpatialVector& rNormal2=m_pNormalArray[arrayIndex2];

		point=Spline<SpatialVector,Real>::Cardinal2D(weight1,
				rVertexM,rVertex0,rVertex1,rVertex2);
		norm=unitSafe(Spline<SpatialVector,Real>::Cardinal2D(weight1,
				rNormalM,rNormal0,rNormal1,rNormal2));
	}
	else if(m_sampleMethod==e_pointNormal)
	{
		//* TODO can we cache the configure?

		TrianglePN<Real> trianglePN;
		trianglePN.configure(rVertex0,rVertex1,rVertex1+1e-3*rNormal1,
				rNormal0,rNormal1,rNormal1);
		SpatialBary segmentBary(weight0,weight1,0.0);
		trianglePN.solve(segmentBary,point,norm);
	}
	else
	{
		point=rVertex0*weight0+rVertex1*weight1;
		norm=rNormal0*weight0+rNormal1*weight1;
	}

	const Real mag=magnitude(norm);
	SpatialTransform result;
	if(mag>0.0)
	{
		SpatialVector tangent=a_tangent;

		if(m_pUVArray)
		{
			const SpatialVector& rUV0=m_pUVArray[arrayIndex];
			const SpatialVector& rUV1=m_pUVArray[arrayIndex+1];

			SpatialVector du;
			SpatialVector dv;
			triangleDuDv(du,dv,rVertex0,rVertex1,rVertex1,rUV0,rUV1,rUV1);

			tangent=du;

#if FE_SC_SAMPLE_DEBUG
			feLog("du %s dv %s\n",c_print(du),c_print(dv));
#endif
		}
		if(magnitudeSquared(tangent)<1e-6)
		{
			tangent=rVertex1-rVertex0;
		}
		if(magnitudeSquared(tangent)<1e-12)
		{
			feX("SurfaceCurves::sample","coincident consecutive vertices");
		}
		makeFrameNormalY(result,point,tangent,norm/mag);

#if FE_SC_SAMPLE_DEBUG
		feLog("a_tangent %s tangent %s\n",c_print(a_tangent),c_print(tangent));
#endif
	}
	else
	{
		setIdentity(result);
		setTranslation(result,point);
	}

#if FE_SC_SAMPLE_DEBUG
	feLog("vertices %d start %d+%d\n",m_vertices,startIndex,subCount);
	feLog("realIndex %.6G subIndex %d/%d weight0 %.6G weight1 %.6G\n",
			realIndex,subIndex,subCount,weight0,weight1);
	feLog("arrayIndex %d %d\n",arrayIndex,arrayIndex1);
	feLog("rVertex0 %s\n",c_print(rVertex0));
	feLog("rVertex1 %s\n",c_print(rVertex1));
	feLog("rNormal0 %s\n",c_print(rNormal0));
	feLog("rNormal1 %s\n",c_print(rNormal1));
	feLog("point %s\n",c_print(point));
	feLog("norm %s\n",c_print(norm));
	feLog("result %s\n",c_print(result));
#endif

	return result;
}


SpatialTransform SurfaceCurves::sample(Vector2 a_uv) const
{
	SpatialTransform transform;

	//* TODO
	setIdentity(transform);

	return transform;
}

void SurfaceCurves::Impact::draw(const SpatialTransform& a_rTransform,
	sp<DrawI> a_spDrawI,const fe::Color* a_pColor) const
{
	SpatialVector line[2];
	line[0]=m_vertex0;
	line[1]=m_vertex1;

	a_spDrawI->drawTransformedLines(a_rTransform,NULL,
			line,NULL,2,DrawI::e_strip,false,a_pColor);
}

void SurfaceCurves::draw(const SpatialTransform& a_transform,
	sp<DrawI> a_spDrawI,const Color* color) const
{
	for(U32 elementIndex=0;elementIndex<m_elements;elementIndex++)
	{
		const U32 start=m_pElementArray[elementIndex][0];
		const U32 subCount=m_pElementArray[elementIndex][1];
		const BWORD openCurve=m_pElementArray[elementIndex][2];
		if(!openCurve)
		{
			//* TODO something
			continue;
		}

		a_spDrawI->drawTransformedLines(a_transform,NULL,
				&m_pVertexArray[start],
				m_pNormalArray? &m_pNormalArray[start]: NULL,
				subCount,DrawI::e_strip,false,color);
	}
}

void SurfaceCurves::resolveImpact(sp<ImpactI> a_spImpactI) const
{
	sp<Impact> spImpact(a_spImpactI);

	// TEMP if no nearest
	if(!spImpact.isValid() || spImpact->distance()<Real(0))
	{
		return;
	}

	SpatialVector normal;
	set(normal,1.0,0.0,0.0);	//* TODO

	spImpact->setNormalLocal(normal);
}

} // namespace
